import os
import threading
from datetime import datetime, timezone
from random import randrange

_local = threading.local()


def _get_increment() -> int:
    if not hasattr(_local, "increment"):
        _local.increment = randrange(0, 4095)

    increment = _local.increment
    _local.increment += 1
    return increment


class Snowflake:
    """A Snowflake ID (https://en.wikipedia.org/wiki/Snowflake_ID).
    This class wraps an integer and adds convenience functions."""

    EPOCH = 1_640_995_200_000  # 2022-01-01 at 00:00:00 UTC

    _raw: int

    def __init__(self, src: int):
        self._raw = src

    def __str__(self) -> str:
        return str(self.id)

    def __repr__(self) -> str:
        return f"Snowflake<{self.id}, {self.process}, {self.thread}, {self.increment}, {self.timestamp}>"

    def __int__(self) -> int:
        return self._raw

    def __float__(self) -> float:
        return float(self._raw)

    def __lt__(self, y: "Snowflake"):
        return self.id < y.id

    def __le__(self, y: "Snowflake"):
        return self.id <= y.id

    def __eq__(self, y: "Snowflake"):
        return self.id == y.id

    def __ne__(self, y: "Snowflake"):
        return self.id != y.id

    def __gt__(self, y: "Snowflake"):
        return self.id > y.id

    def __ge__(self, y: "Snowflake"):
        return self.id >= y.id

    @property
    def id(self) -> int:
        """The raw integer value of the snowflake."""
        return self._raw

    @property
    def time(self) -> datetime:
        """The time embedded into the snowflake."""
        return datetime.fromtimestamp(self.timestamp, tz=timezone.utc)

    @property
    def timestamp(self) -> float:
        """The unix timestamp embedded into the snowflake."""
        return ((self._raw >> 22) + self.EPOCH) / 1000

    @property
    def process(self) -> int:
        """The process ID embedded into the snowflake."""
        return (self._raw & 0x3E0000) >> 17

    @property
    def thread(self) -> int:
        """The thread ID embedded into the snowflake."""
        return (self._raw & 0x1F000) >> 12

    @property
    def increment(self) -> int:
        """The increment embedded into the snowflake."""
        return self._raw & 0xFFF

    @classmethod
    def generate(cls, time: datetime | None = None):
        """Generates a new snowflake.
        If `time` is set, use that time for the snowflake, otherwise, use the current time.
        """

        process_id = os.getpid()
        thread_id = threading.get_native_id()
        increment = _get_increment()
        now = time if time else datetime.now(tz=timezone.utc)
        timestamp = round(now.timestamp() * 1000) - cls.EPOCH

        return cls(
            timestamp << 22
            | (process_id % 32) << 17
            | (thread_id % 32) << 12
            | (increment % 4096)
        )

    @classmethod
    def generate_int(cls, time: datetime | None = None):
        return cls.generate(time).id
