from datetime import datetime
import enum

from sqlalchemy import Text, Integer, BigInteger, ForeignKey, DateTime
from sqlalchemy.dialects.postgresql import ARRAY
from sqlalchemy.orm import Mapped, mapped_column, relationship

from .base import Base
from .snowflake import Snowflake


class User(Base):
    __tablename__ = "users"

    id: Mapped[int] = mapped_column(
        BigInteger(), primary_key=True, default=Snowflake.generate_int
    )
    username: Mapped[str] = mapped_column(Text(), unique=True, nullable=False)
    display_name: Mapped[str | None] = mapped_column(Text(), nullable=True)
    bio: Mapped[str | None] = mapped_column(Text(), nullable=True)

    tokens: Mapped[list["Token"]] = relationship(
        back_populates="user", cascade="all, delete-orphan"
    )
    auth_methods: Mapped[list["AuthMethod"]] = relationship(
        back_populates="user", cascade="all, delete-orphan"
    )

    def __repr__(self):
        return f"User(id={self.id!r}, username={self.username!r})"


class Token(Base):
    __tablename__ = "tokens"

    id: Mapped[int] = mapped_column(
        BigInteger(), primary_key=True, default=Snowflake.generate_int
    )
    expires_at: Mapped[datetime] = mapped_column(DateTime(), nullable=False)
    scopes: Mapped[list[str]] = mapped_column(ARRAY(Text), nullable=False)

    user_id: Mapped[int] = mapped_column(ForeignKey("users.id"))
    user: Mapped[User] = relationship(back_populates="tokens")

    def __repr__(self):
        return f"Token(id={self.id!r}, user={self.user_id!r})"

    def has_scope(self, scope: str):
        """Returns whether this token can be used for the given scope."""

        # `*` is a special scope for site tokens, which grants access to all endpoints.
        if "*" in self.scopes:
            return True

        # Some scopes have sub-scopes, indicated by a `.` (i.e. `user.edit` is contained in `user`)
        # Tokens can have these narrower scopes given to them, or the wider, more privileged scopes
        # This way, both `user` and `user.edit` tokens will grant access to `user.edit` endpoints.
        return scope in self.scopes or scope.split(".")[0] in self.scopes


class AuthType(enum.IntEnum):
    DISCORD = 1
    GOOGLE = 2
    TUMBLR = 3
    FEDIVERSE = 4
    EMAIL = 5


class AuthMethod(Base):
    __tablename__ = "auth_methods"

    id: Mapped[int] = mapped_column(
        BigInteger(), primary_key=True, default=Snowflake.generate_int
    )
    auth_type: Mapped[AuthType] = mapped_column(Integer(), nullable=False)

    remote_id: Mapped[str] = mapped_column(Text(), nullable=False)
    remote_username: Mapped[str | None] = mapped_column(Text(), nullable=True)

    user_id: Mapped[int] = mapped_column(ForeignKey("users.id"))
    user: Mapped[User] = relationship(back_populates="auth_methods")

    fediverse_app_id: Mapped[int] = mapped_column(
        ForeignKey("fediverse_apps.id"), nullable=True
    )
    fediverse_app: Mapped["FediverseApp"] = relationship()


class FediverseInstanceType(enum.IntEnum):
    MASTODON_API = 1
    MISSKEY_API = 2


class FediverseApp(Base):
    __tablename__ = "fediverse_apps"

    id: Mapped[int] = mapped_column(
        BigInteger(), primary_key=True, default=Snowflake.generate_int
    )
    instance: Mapped[str] = mapped_column(Text(), unique=True, nullable=False)
    client_id: Mapped[str] = mapped_column(Text(), nullable=False)
    client_secret: Mapped[str] = mapped_column(Text(), nullable=False)
    instance_type: Mapped[FediverseInstanceType] = mapped_column(
        Integer(), nullable=False
    )
